"""
Utils module for common test functionality.
"""
import os


def mkfile(tmpdir, filename, content=""):
    """
    A simple function for creating a file and returning the path. This is to
    abstract out file creation logic in the tests.

    :param tmpdir: (str) The path to the temp directory.
    :param filename: (str) The name of the file.
    :param contents: (str) Optional contents to write to the file. Defaults to
        an empty string.
    :returns: (str) The appended path of the given tempdir and filename.
    """
    filepath = os.path.join(tmpdir, filename)

    with open(filepath, "w") as f:
        f.write(content)

    return filepath



#!/usr/bin/env python2
"""This module will collect information on flux jobs from the live kvs
store and output times for each phase.

create: The time flux registerd the job
starting: The time the job was created
running: The time the job was running
completing: The time the job started its' completion pahse.
complete: The time the job was complete
walltime: ? Seems to be 0.
"""
import os

import flux
from flux import kvs


f = flux.Flux()

fs = "FLUX_START_SECONDS"
if fs in os.environ:
    print("flux start: {0}".format(os.environ[fs]))

for d in kvs.walk("lwj", flux_handle=f):
    try:
        # print(type(d))
        fdir = "lwj.{0}".format(d[0])

        qcreate = "{0}.create-time".format(fdir)
        create_time = kvs.get(f, qcreate)

        qstart = "{0}.starting-time".format(fdir)
        start_time = kvs.get(f, qstart)

        qrun = "{0}.running-time".format(fdir)
        start_time = kvs.get(f, qrun)

        qcomplete = "{0}.complete-time".format(fdir)
        complete_time = kvs.get(f, qcomplete)

        qcompleting = "{0}.completing-time".format(fdir)
        completing_time = kvs.get(f, qcompleting)

        qwall = "{0}.walltime".format(fdir)
        wall_time = kvs.get(f, qwall)

        proto = "Job {0}: create: {1} start {1} run {2} completing {3} complete {4} wall {5}"
        print(
            proto.format(
                d[0], create_time, start_time, completing_time, complete_time, wall_time
            )
        )
    except BaseException:
        pass


# vi: ts=4 sw=4 expandtab



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""Updated celery configuration."""
from __future__ import (
    absolute_import,
    print_function,
)

import logging
import os

import billiard
import psutil
from celery import Celery
from celery.signals import worker_process_init

import merlin.common.security.encrypt_backend_traffic
from merlin.config import (
    broker,
    results_backend,
)
from merlin.log_formatter import FORMATS
from merlin.router import route_for_task


LOG = logging.getLogger(__name__)

merlin.common.security.encrypt_backend_traffic.set_backend_funcs()

broker_ssl = True
results_ssl = False
try:
    BROKER_URI = broker.get_connection_string()
    LOG.info(f"broker: {broker.get_connection_string(include_password=False)}")
    broker_ssl = broker.get_ssl_config()
    LOG.info(f"broker_ssl = {broker_ssl}")
    RESULTS_BACKEND_URI = results_backend.get_connection_string()
    results_ssl = results_backend.get_ssl_config(celery_check=True)
    LOG.info(
        f"results: {results_backend.get_connection_string(include_password=False)}"
    )
    LOG.info(f"results: redis_backed_use_ssl = {results_ssl}")
except ValueError:
    # These variables won't be set if running with '--local'.
    BROKER_URI = None
    RESULTS_BACKEND_URI = None


app = Celery(
    "merlin",
    broker=BROKER_URI,
    backend=RESULTS_BACKEND_URI,
    broker_use_ssl=broker_ssl,
    redis_backend_use_ssl=results_ssl,
)


app.conf.update(
    task_serializer="pickle", accept_content=["pickle"], result_serializer="pickle"
)

app.autodiscover_tasks(["merlin.common"])

app.conf.update(
    task_acks_late=True,
    task_reject_on_worker_lost=True,
    task_publish_retry_policy={
        "interval_start": 10,
        "interval_step": 10,
        "interval_max": 60,
    },
    redis_max_connections=100000,
)

# Set a one hour timeout to acknowledge a task before it's available to grab
# again.
app.conf.broker_transport_options = {"visibility_timeout": 7200, "max_connections": 100}

app.conf.update(broker_pool_limit=0)

# Task routing: call our default queue merlin
app.conf.task_routes = (route_for_task,)
app.conf.task_default_queue = "merlin"

# Log formatting
app.conf.worker_log_color = True
app.conf.worker_log_format = FORMATS["DEFAULT"]
app.conf.worker_task_log_format = FORMATS["WORKER"]


@worker_process_init.connect()
def setup(**kwargs):
    """
    Set affinity for the worker on startup (works on toss3 nodes)

    :param `**kwargs`: keyword arguments
    """
    if "CELERY_AFFINITY" in os.environ and int(os.environ["CELERY_AFFINITY"]) > 1:
        # Number of cpus between workers.
        cpu_skip = int(os.environ["CELERY_AFFINITY"])
        npu = psutil.cpu_count()
        p = psutil.Process()
        current = billiard.current_process()
        prefork_id = current._identity[0] - 1  # range 0:nworkers-1
        cpu_slot = (prefork_id * cpu_skip) % npu
        p.cpu_affinity(list(range(cpu_slot, cpu_slot + cpu_skip)))



import argparse
import sys

from faker import Faker


fake = Faker()


def process_args(args):
    n_samples = args.n
    # n_batches = args.b
    outfile = args.outfile

    with open(outfile, "w") as name_file:
        for samp in range(n_samples):
            name_file.write("{}\n".format(fake.first_name_male()))

        # for batch in range(n_batches):
        #     for samp in range(n_samples):
        #         name_file.write('{}, {}\n'.format(batch, fake.first_name_male()))
        # name_batch = ','.join([fake.first_name_male() for samp in range(n_samples)])
        # name_file.write('{}\n'.format(name_batch))


def setup_argparse():
    parser = argparse.ArgumentParser("Generate some names!")
    parser.add_argument("-n", help="number of names", default=100, type=int)
    parser.add_argument("-b", help="number of batches of names", default=5, type=int)
    parser.add_argument(
        "-outfile", help="name of output .csv file", default="samples.csv"
    )
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    process_args(args)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import json
import sys


def process_args(args):
    result = []
    for path in args.instring.split("\n"):
        with open(path, "r") as json_file:
            result.append(json.load(json_file))

    with open(args.outfile, "w") as outfile:
        json.dump(result, outfile)


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="Convert a list of json files into numpy"
    )

    parser.add_argument(
        "-instring", help="text containing list of json files", default=""
    )
    parser.add_argument("-outfile", help="json file", default="results.json")
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    process_args(args)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import os
import pathlib
import sys

import pandas as pd


def load_samples(sample_file_paths):
    """Read in processed sample files, returning flat list."""
    samples = []
    for sample_file_path in sample_file_paths:
        with open(sample_file_path, "r") as sample_file:
            for sample in sample_file:
                sample = sample.strip()
                samples.append(sample)

    return samples


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="Read in and analyze samples from plain text file"
    )

    parser.add_argument(
        "sample_file_paths", help="paths to sample files", default="", nargs="+"
    )

    parser.add_argument(
        "--results", help="Name of output json file", default="samples.json"
    )

    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()

    # Collect the samples
    samples = load_samples(args.sample_file_paths)

    # Count up the occurences
    namesdf = pd.DataFrame({"Name": samples})

    names = namesdf["Name"].value_counts()

    # Serialize processed samples
    #  create directory if it doesn't exist already
    abspath = os.path.abspath(args.results)
    absdir = os.path.dirname(abspath)
    if not os.path.isdir(absdir):
        pathlib.Path(absdir).mkdir(parents=True, exist_ok=True)

    names.to_json(args.results)


if __name__ == "__main__":
    sys.exit(main())



import logging
import re

import yaml


LOG = logging.getLogger(__name__)


def error_override_vars(override_vars, spec_filepath):
    """
    Warn user if any given variable name isn't found in the original spec file.
    """
    if override_vars is None:
        return
    original_text = open(spec_filepath, "r").read()
    for variable in override_vars.keys():
        if variable not in original_text:
            raise ValueError(
                f"Command line override variable '{variable}' not found in spec file '{spec_filepath}'."
            )


def replace_override_vars(full_text, env, override_vars):
    """
    Given the full text of a yaml spec, return the full
    text with user variable overrides in the 'env' block.

    The env yaml block looks like:

        env:
            variables:
                .......
                .......

    The regex will find and match to the above yaml block.
    """
    updated_env = dict(env)
    if override_vars is not None:
        for key, val in env.items():
            updated_env[key].update(override_vars)
    updated_env = {"env": updated_env}
    dump = yaml.dump(updated_env, default_flow_style=False, sort_keys=False)
    updated_env_text = f"\n{dump}\n"

    env_block_pattern = r"\nenv\s*:\s*(\n+( |\t)+.*)+\n*"
    regex = re.compile(env_block_pattern)
    updated_full_text = re.sub(regex, updated_env_text, full_text)
    return updated_full_text


def dump_with_overrides(spec, override_vars):
    dumped_text = spec.dump()
    if override_vars is None:
        return dumped_text
    result = replace_override_vars(
        full_text=dumped_text, env=spec.environment, override_vars=override_vars
    )
    return result



import argparse
import sys

import numpy as np


try:
    import cPickle as pickle
except ImportError:
    import pickle


def setup_argparse():
    parser = argparse.ArgumentParser(description="Use a regressor to make a prediction")

    parser.add_argument(
        "-infile", help=".npy file with data to predict", default="new_x.npy"
    )
    parser.add_argument(
        "-reg", help="pickled regressor file", default="random_forest_reg.pkl"
    )
    parser.add_argument(
        "-outfile", help="file to store the new predictions", default="new_y.npy"
    )

    return parser


def predict(args):
    regr = pickle.load(open(args.reg, "rb"))

    X = np.load(args.infile)

    new_y = regr.predict(X)
    np.save(open(args.outfile, "wb"), new_y)


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    predict(args)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import ast
import sys

import numpy as np

from merlin.common.util_sampling import scale_samples


def process_args(args):
    n_samples = args.n
    n_dims = args.dims
    sample_type = args.sample_type
    if sample_type == "random":
        x = np.random.random((n_samples, n_dims))
    elif sample_type == "grid":
        subdivision = int(pow(n_samples, 1 / float(n_dims)))
        temp = [np.linspace(0, 1.0, subdivision) for i in range(n_dims)]
        X = np.meshgrid(*temp)
        x = np.stack([xx.flatten() for xx in X], axis=1)

    if args.scale is not None:
        limits = []
        do_log = []
        scales = ast.literal_eval(args.scale)
        for scale in scales:
            limits.append((scale[0], scale[1]))
            if scale[2] == "log":
                do_log.append(True)
            else:
                do_log.append(False)
        x = scale_samples(x, limits, do_log=do_log)

    np.save(args.outfile, x)


def setup_argparse():
    parser = argparse.ArgumentParser("Generate some samples!")
    parser.add_argument("-n", help="number of samples", default=100, type=int)
    parser.add_argument("-dims", help="number of dimensions", default=2, type=int)
    parser.add_argument(
        "-sample_type",
        help="type of sampling. options: random, grid. If grid, will try to get close to the correct number of samples",
        default="random",
    )
    parser.add_argument(
        "-scale",
        help='ranges to scale results in form "[(min,max,type),(min, max,type)]" where type = "linear" or "log"',
    )
    parser.add_argument("-outfile", help="name of output .npy file", default="samples")
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    process_args(args)


if __name__ == "__main__":
    sys.exit(main())



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
Module for project-wide utility functions.
"""
import getpass
import logging
import os
import re
import socket
import subprocess
from contextlib import (
    contextmanager,
    suppress,
)
from copy import deepcopy
from types import SimpleNamespace

import numpy as np
import psutil
import yaml


try:
    import cPickle as pickle
except ImportError:
    import pickle


LOG = logging.getLogger(__name__)
ARRAY_FILE_FORMATS = ".npy, .csv, .tab"
DEFAULT_FLUX_VERSION = "0.13"


def get_user_process_info(user=None, attrs=None):
    """
    Return a list of process info for all of the user's running processes.

    :param `user`: user name (default from getpass). Option: 'all_users': get
        all processes
    :param `atts`: the attributes to include
    """
    if attrs is None:
        attrs = ["pid", "name", "username", "cmdline"]

    if "username" not in attrs:
        attrs.extend("username")

    if user is None:
        user = getpass.getuser()

    if user == "all_users":
        return [p.info for p in psutil.process_iter(attrs=attrs)]
    else:
        return [
            p.info
            for p in psutil.process_iter(attrs=attrs)
            if user in p.info["username"]
        ]


def check_pid(pid, user=None):
    """
    Check if pid is in process list.

    :param `pid`: process id
    :param `user`: user name (default from getpass). Option: 'all_users': get
        all processes
    """
    user_processes = get_user_process_info(user=user)
    for p in user_processes:
        if int(p["pid"]) == pid:
            return True
    return False


def get_pid(name, user=None):
    """
    Return pid of process with name.

    :param `name`: process name
    :param `user`: user name (default from getpass). Option: 'all_users': get
        all processes
    """
    user_processes = get_user_process_info(user=user)
    name_list = [p["pid"] for p in user_processes if name in p["name"]]
    if name_list:
        return name_list
    return None


def get_procs(name, user=None):
    """
    Return a list of (pid, cmdline) tuples of process with name.

    :param `name`: process name
    :param `user`: user name (default from getpass). Option: 'all_users': get
        all processes
    """
    user_processes = get_user_process_info(user=user)
    procs = [(p["pid"], p["cmdline"]) for p in user_processes if name in p["name"]]
    return procs


def is_running_psutil(cmd, user=None):
    """
    Determine if process with given command is running.
    Uses psutil command instead of call to 'ps'

    :param `cmd`: process cmd
    :param `user`: user name (default from getpass). Option: 'all_users': get
        all processes
    """
    user_processes = get_user_process_info(user=user)
    return any(cmd in " ".join(p["cmdline"]) for p in user_processes)


def is_running(name, all_users=False):
    """
    Determine if process with name is running.

    :param `name`: process name
    """
    cmd = ["ps", "ux"]

    if all_users:
        cmd[1] = "aux"

    try:
        ps = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, encoding="utf8"
        ).communicate()[0]
    except TypeError:
        ps = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]

    if name in ps:
        return True

    return False


def expandvars2(path):
    """
    Replace shell strings from the current environment variables

    :param `path`: a path
    """
    return re.sub(r"(?<!\\)\$[A-Za-z_][A-Za-z0-9_]*", "", os.path.expandvars(path))


def regex_list_filter(regex, list_to_filter, match=True):
    """
    Apply a regex filter to a list

    :param `regex`          : the regular expression
    :param `list_to_filter` : the list to filter

    :return `new_list`
    """
    r = re.compile(regex)
    if match:
        return list(filter(r.match, list_to_filter))
    return list(filter(r.search, list_to_filter))


def load_yaml(filepath):
    """
    Safely read a yaml file.

    :param `filepath`: a filepath to a yaml file
    :type filepath: str

    :returns: Python objects holding the contents of the yaml file
    """
    with open(filepath, "r") as _file:
        return yaml.safe_load(_file)


def get_yaml_var(entry, var, default):
    """
    Return entry[var], else return default

    :param `entry`: a yaml dict
    :param `var`: a yaml key
    :param `default`: default value in the absence of data
    """
    ret = default

    if isinstance(entry, dict):
        with suppress(KeyError):
            ret = entry[var]
    else:
        with suppress(AttributeError):
            ret = getattr(entry, var)
    return ret


def load_array_file(filename, ndmin=2):
    """
    Loads up an array stored in filename, based on extension.

    Valid filename extensions:
        '.npy'  :  numpy binary file
        '.csv'  :  comma separated text file
        '.tab'  :  whitespace (or tab) separated text file

    :param `filename` : The file to load
    :param `ndmin`    : The minimum number of dimensions to load
    """

    protocol = determine_protocol(filename)

    # Don't change binary-stored numpy arrays; just check dimensions
    if protocol == "npy":
        array = np.load(filename)
        if array.ndim < ndmin:
            LOG.error(
                f"Array in {filename} has fewer than the required \
                       minimum dimensions ({array.ndim} < {ndmin})!"
            )
    # Make sure text files load as strings with minimum number of dimensions
    elif protocol == "csv":
        array = np.loadtxt(filename, delimiter=",", ndmin=ndmin, dtype=np.str)
    elif protocol == "tab":
        array = np.loadtxt(filename, ndmin=ndmin, dtype=np.str)
    else:
        raise TypeError(
            f"{protocol} is not a valid array file extension.\
                         Choices: {ARRAY_FILE_FORMATS}"
        )

    return array


def determine_protocol(fname):
    """
    Determines a file protocol based on file name extension.
    """
    _, ext = os.path.splitext(fname)
    if ext.startswith("."):
        protocol = ext.lower().strip(".")
    else:
        raise ValueError(f"{fname} needs an ext (eg .hdf5) to determine protocol!")
    # Map .h5 to .hdf5
    if protocol == "h5":
        protocol = "hdf5"
    return protocol


@contextmanager
def cd(path):
    """
    TODO
    """
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def pickle_data(filepath, content):
    """Dump content to a pickle file"""
    with open(filepath, "w") as f:
        pickle.dump(content, f)


def get_source_root(filepath):
    """Used to find the absolute project path given a sample file path from
    within the project.
    """
    filepath = os.path.abspath(filepath)
    sep = os.path.sep
    if filepath == sep:
        return None

    parent = os.path.dirname(filepath)
    # Walk backwards testing for integers.
    for _, _dir in enumerate(parent.split(sep)[::-1]):
        try:
            int(_dir)
        except ValueError:
            break_point = _dir
            break

    root, _ = parent.split(break_point)
    root = os.path.split(root)[0]
    return root


def ensure_directory_exists(**kwargs):
    """
    TODO
    """
    aggregate_bundle = kwargs["aggregate_file"]
    dirname = os.path.dirname(aggregate_bundle)

    if not os.path.exists(dirname):
        LOG.info(f"making directories to {dirname}.")
        os.makedirs(dirname)
        return False
    return True


def nested_dict_to_namespaces(dic):
    """Code for recursively converting dictionaries of dictionaries
        into SimpleNamespaces instead.
    """

    def recurse(dic):
        if not isinstance(dic, dict):
            return dic
        for key, val in list(dic.items()):
            dic[key] = recurse(val)
        return SimpleNamespace(**dic)

    if not isinstance(dic, dict):
        raise TypeError(f"{dic} is not a dict")

    new_dic = deepcopy(dic)
    return recurse(new_dic)


def nested_namespace_to_dicts(ns):
    """Code for recursively converting namespaces of namespaces
        into dictionaries instead.
    """

    def recurse(ns):
        if not isinstance(ns, SimpleNamespace):
            return ns
        for key, val in list(ns.__dict__.items()):
            setattr(ns, key, recurse(val))
        return ns.__dict__

    if not isinstance(ns, SimpleNamespace):
        raise TypeError(f"{ns} is not a SimpleNamespace")

    new_ns = deepcopy(ns)
    return recurse(new_ns)


def get_flux_version(flux_path, no_errors=False):
    """
    Return the flux version as a string

    :param `flux_path`: the full path to the flux bin
    :param `no_errors`: a flag to determine if this a test run to ignore errors
    """
    cmd = [flux_path, "version"]

    ps = None

    try:
        ps = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, encoding="utf8"
        ).communicate()
    except FileNotFoundError as e:
        if not no_errors:
            LOG.error(f"The flux path {flux_path} canot be found")
            raise e

    try:
        flux_ver = re.search(r"\s*([\d.]+)", ps[0]).group(1)
    except (ValueError, TypeError) as e:
        if not no_errors:
            LOG.error(f"The flux version canot be determined")
            raise e
        else:
            flux_ver = DEFAULT_FLUX_VERSION
            LOG.warning(f"Using syntax for default version: {flux_ver}")

    return flux_ver


def get_flux_cmd(flux_path, no_errors=False):
    """
    Return the flux command as string

    :param `flux_path`: the full path to the flux bin
    :param `no_errors`: a flag to determine if this a test run to ignore errors
    """
    # The default is for flux version >= 0.13,
    # this may change in the future.
    flux_cmd = "flux mini run"

    flux_ver = get_flux_version(flux_path, no_errors=no_errors)

    vers = [int(n) for n in flux_ver.split(".")]
    if vers[0] == 0 and vers[1] < 13:
        flux_cmd = "flux wreckrun"

    return flux_cmd


def check_machines(machines):
    """
    Return a True if the current machine is in the list of machines.

    :param `machines`: A single machine or list of machines to compare
                       with the current machine.
    """
    local_hostname = socket.gethostname()

    if not isinstance(machines, (list, tuple)):
        machines = [machines]

    for mach in machines:
        if mach in local_hostname:
            return True

    return False



import argparse
import json
import sys


def process_args(args):
    results = {
        "inputs": {"X": args.X, "Y": args.Y, "Z": args.Z},
        "outputs": {
            "X+Y+Z": args.X + args.Y + args.Z,
            "X*Y*Z": args.X * args.Y * args.Z,
        },
    }

    with open(args.outfile, "w") as f:
        json.dump(results, f)


def setup_argparse():
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument(
        "X", metavar="X", type=float, help="The x dimension of the sample."
    )
    parser.add_argument(
        "Y", metavar="Y", type=float, help="The y dimension of the sample."
    )
    parser.add_argument(
        "Z", metavar="Z", type=float, help="The z dimension of the sample."
    )
    parser.add_argument(
        "-outfile", help="Output file name", default="hello_world_output.json"
    )
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    process_args(args)


if __name__ == "__main__":
    sys.exit(main())



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.0.5.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
Script for running Merlin command line interface tests.
Built for 1) manual use and 2) continuous integration.
"""
import argparse
import os
import shutil
import sys
import time
from contextlib import suppress
from glob import glob
from re import search
from subprocess import (
    PIPE,
    Popen,
)

from merlin.utils import get_flux_cmd


OUTPUT_DIR = "cli_test_studies"


def run_single_test(name, test, test_label="", buffer_length=50):
    dot_length = buffer_length - len(name) - len(str(test_label))
    print(f"TEST {test_label}: {name}{'.'*dot_length}", end="")
    command = test[0]
    conditions = test[1]
    if not isinstance(conditions, list):
        conditions = [conditions]

    start_time = time.time()
    process = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
    stdout, stderr = process.communicate()
    end_time = time.time()
    total_time = end_time - start_time
    if stdout is not None:
        stdout = stdout.decode("utf-8")
    if stderr is not None:
        stderr = stderr.decode("utf-8")
    return_code = process.returncode

    info = {
        "total_time": total_time,
        "command": command,
        "stdout": stdout,
        "stderr": stderr,
        "return_code": return_code,
    }

    # ensure all test conditions are satisfied
    for condition in conditions:
        condition.ingest_info(info)
        passed = condition.passes
        if passed is False:
            break

    return passed, info


def clear_test_studies_dir():
    """
    Deletes the 'test_studies' directory, in order to preserve
    state each time cli tests are run.
    """
    with suppress(FileNotFoundError):
        shutil.rmtree(f"./{OUTPUT_DIR}")


def process_test_result(passed, info, is_verbose, exit):
    """
    Process and print test results to the console.
    """
    # if the environment does not contain necessary programs, exit early.
    if passed is False and "merlin: command not found" in info["stderr"]:
        print(f"\nMissing from environment:\n\t{info['stderr']}")
        return None
    elif passed is False:
        print("FAIL")
        if exit is True:
            return None
    else:
        print("pass")

    if is_verbose is True:
        print(f"\tcommand: {info['command']}")
        print(f"\telapsed time: {round(info['total_time'], 2)} s")
        if info["return_code"] != 0:
            print(f"\treturn code: {info['return_code']}")
        if info["stderr"] != "":
            print(f"\tstderr:\n{info['stderr']}")

    return passed


def run_tests(args, tests):
    """
    Run all inputted tests.
    :param `tests`: a dictionary of
        {"test_name" : ("test_command", [conditions])}
    """
    selective = False
    n_to_run = len(tests)
    if args.ids is not None and len(args.ids) > 0:
        if not all(x > 0 for x in args.ids):
            raise ValueError(f"Test ids must be between 1 and {len(tests)}, inclusive.")
        selective = True
        n_to_run = len(args.ids)
    elif args.local is not None:
        args.ids = []
        n_to_run = 0
        selective = True
        test_id = 1
        for _, test in tests.items():
            if len(test) == 3 and test[2] == "local":
                args.ids.append(test_id)
                n_to_run += 1
            test_id += 1

    print(f"Running {n_to_run} integration tests...")
    start_time = time.time()

    total = 0
    failures = 0
    for test_name, test in tests.items():
        test_label = total + 1
        if selective and test_label not in args.ids:
            total += 1
            continue
        try:
            passed, info = run_single_test(test_name, test, test_label)
        except BaseException:
            passed = False

        result = process_test_result(passed, info, args.verbose, args.exit)
        if result is None:
            print("Exiting early")
            return 1
        if result is False:
            failures += 1
        total += 1

    end_time = time.time()
    total_time = end_time - start_time

    if failures == 0:
        print(f"Done. {n_to_run} tests passed in {round(total_time, 2)} s.")
        return 0
    print(
        f"Done. {failures} tests out of {n_to_run} failed after {round(total_time, 2)} s.\n"
    )
    return 1


class Condition:
    def ingest_info(self, info):
        """
        This function allows child classes of Condition
        to take in data AFTER a test is run.
        """
        for key, val in info.items():
            setattr(self, key, val)

    @property
    def passes(self):
        print("Extend this class!")
        return False


class ReturnCodeCond(Condition):
    """
    A condition that some process must return 0
    as its return code.
    """

    def __init__(self, expected_code=0):
        """
        :param `expected_code`: the expected return code
        """
        self.expected_code = expected_code

    @property
    def passes(self):
        return self.return_code == self.expected_code


class NoStderrCond(Condition):
    """
    A condition that some process have an empty
    stderr string.
    """

    @property
    def passes(self):
        return self.stderr == ""


class RegexCond(Condition):
    """
    A condition that some body of text MUST match a
    given regular expression. Defaults to stdout.
    """

    def __init__(self, regex, negate=False):
        """
        :param `regex`: a string regex pattern
        """
        self.regex = regex
        self.negate = negate

    def is_within(self, text):
        """
        :param `text`: text in which to search for a regex match
        """
        return search(self.regex, text) is not None

    @property
    def passes(self):
        if self.negate:
            return not self.is_within(self.stdout)
        return self.is_within(self.stdout) or self.is_within(self.stderr)


class StudyCond(Condition):
    """
    An abstract condition that is aware of a study's name and output path.
    """

    def __init__(self, study_name, output_path):
        """
        :param `study_name`: the name of a study
        :param `output_path`: the $(OUTPUT_PATH) of a study
        """
        self.study_name = study_name
        self.output_path = output_path
        self.dirpath_glob = f"{self.output_path}/{self.study_name}" f"_[0-9]*-[0-9]*"

    def glob(self, glob_string):
        candidates = glob(glob_string)
        if isinstance(candidates, list):
            return sorted(candidates)[-1]
        return candidates


class StepFileExistsCond(StudyCond):
    """
    A StudyCond that checks for a particular file's existence.
    """

    def __init__(self, step, filename, study_name, output_path):
        """
        :param `step`: the name of a step
        :param `filename`: name of file to search for in step's workspace directory
        :param `study_name`: the name of a study
        :param `output_path`: the $(OUTPUT_PATH) of a study
        """
        super().__init__(study_name, output_path)
        self.step = step
        self.filename = filename

    def file_exists(self):
        glob_string = f"{self.dirpath_glob}/{self.step}/{self.filename}"
        try:
            filename = self.glob(glob_string)
        except IndexError:
            return False
        return os.path.isfile(filename)

    @property
    def passes(self):
        return self.file_exists()


class StepFileContainsCond(StudyCond):
    """
    A StudyCond that checks that a particular file contains a regex.
    """

    def __init__(self, step, filename, study_name, output_path, regex):
        """
        :param `step`: the name of a step
        :param `filename`: name of file to search for in step's workspace directory
        :param `study_name`: the name of a study
        :param `output_path`: the $(OUTPUT_PATH) of a study
        """
        super().__init__(study_name, output_path)
        self.step = step
        self.filename = filename
        self.regex = regex

    def contains(self):
        glob_string = f"{self.dirpath_glob}/{self.step}/{self.filename}"
        try:
            filename = self.glob(glob_string)
            with open(filename, "r") as textfile:
                filetext = textfile.read()
            return self.is_within(filetext)
        except Exception:
            return False

    def is_within(self, text):
        """
        :param `text`: text in which to search for a regex match
        """
        return search(self.regex, text) is not None

    @property
    def passes(self):
        return self.contains()


class ProvenanceCond(RegexCond):
    """
    A condition that a Merlin provenance yaml spec
    MUST contain a given regular expression.
    """

    def __init__(self, regex, name, output_path):
        """
        :param `regex`: a string regex pattern
        :param `name`: the name of a study
        :param `output_path`: the $(OUTPUT_PATH) of a study
        """
        super().__init__(regex)
        self.name = name
        self.output_path = output_path

    def is_within(self):
        """
        Uses glob to find the correct provenance yaml spec.
        Returns True if that file contains a match to this
        object's self.regex string.
        """
        filepath = (
            f"{self.output_path}/{self.name}"
            f"_[0-9]*-[0-9]*/merlin_info/{self.name}.yaml"
        )
        filename = sorted(glob(filepath))[-1]
        with open(filename, "r") as _file:
            text = _file.read()
            return super().is_within(text)

    @property
    def passes(self):
        return self.is_within()


def define_tests():
    """
    Returns a dictionary of tests, where the key
    is the test's name, and the value is a tuple
    of (shell command, condition(s) to satisfy).
    """
    celery_regex = r"(srun\s+.*)?celery\s+worker\s+(-A|--app)\s+merlin\s+.*"

    # shortcut string variables
    workers = "merlin run-workers"
    run = "merlin run"
    restart = "merlin restart"
    purge = "merlin purge"
    examples = "merlin/examples/workflows"
    demo = f"{examples}/feature_demo/feature_demo.yaml"
    simple = f"{examples}/simple_chain/simple_chain.yaml"
    slurm = f"{examples}/slurm/slurm_test.yaml"
    slurm_restart = f"{examples}/slurm/slurm_par_restart.yaml"
    flux = f"{examples}/flux/flux_test.yaml"
    flux_restart = f"{examples}/flux/flux_par_restart.yaml"
    lsf = f"{examples}/lsf/lsf_par.yaml"
    black = "black --check --target-version py36"
    config_dir = "./CLI_TEST_MERLIN_CONFIG"

    return {
        "merlin": ("merlin", ReturnCodeCond(1), "local"),
        "merlin help": ("merlin --help", ReturnCodeCond(), "local"),
        "merlin version": ("merlin --version", ReturnCodeCond(), "local"),
        "merlin config": (
            f"merlin config -o {config_dir}; rm -rf {config_dir}",
            ReturnCodeCond(),
            "local",
        ),
        "run-workers echo simple_chain": (
            f"{workers} {simple} --echo",
            [ReturnCodeCond(), RegexCond(celery_regex)],
            "local",
        ),
        "run-workers echo feature_demo": (
            f"{workers} {demo} --echo",
            [ReturnCodeCond(), RegexCond(celery_regex)],
            "local",
        ),
        "run-workers echo slurm_test": (
            f"{workers} {slurm} --echo",
            [ReturnCodeCond(), RegexCond(celery_regex)],
            "local",
        ),
        "run-workers echo flux_test": (
            f"{workers} {flux} --echo",
            [ReturnCodeCond(), RegexCond(celery_regex)],
            "local",
        ),
        "run-workers echo override feature_demo": (
            f"{workers} {demo} --echo --vars VERIFY_QUEUE=custom_verify_queue",
            [ReturnCodeCond(), RegexCond("custom_verify_queue")],
            "local",
        ),
        "run feature_demo": (f"{run} {demo}", ReturnCodeCond()),
        "purge feature_demo": (f"{purge} {demo} -f", ReturnCodeCond()),
        "dry feature_demo": (
            f"{run} {demo} --local --dry --vars OUTPUT_PATH=./{OUTPUT_DIR}",
            [
                StepFileExistsCond(
                    "verify", "*/verify_*.sh", "feature_demo", OUTPUT_DIR
                ),
                ReturnCodeCond(),
            ],
            "local",
        ),
        "local simple_chain": (
            f"{run} {simple} --local --vars OUTPUT_PATH=./{OUTPUT_DIR}",
            ReturnCodeCond(),
            "local",
        ),
        "example failure": (f"merlin example failure", RegexCond("not found"), "local"),
        "example simple_chain": (
            f"merlin example simple_chain ; {run} simple_chain.yaml --local --vars OUTPUT_PATH=./{OUTPUT_DIR} ; rm simple_chain.yaml",
            ReturnCodeCond(),
            "local",
        ),
        # "restart local simple_chain": (
        #    f"{restart} --local $(find studies/ -type d -name 'simple_chain_*')",
        #    [ReturnCodeCond(), NoStderrCond()],
        # ),
        "dry launch slurm": (
            f"{run} {slurm} --dry --local --no-errors --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR}",
            StepFileContainsCond(
                "runs", "*/runs.slurm.sh", "slurm_test", OUTPUT_DIR, "srun "
            ),
            "local",
        ),
        "dry launch flux": (
            f"{run} {flux} --dry --local --no-errors --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR}",
            StepFileContainsCond(
                "runs",
                "*/runs.slurm.sh",
                "flux_test",
                OUTPUT_DIR,
                get_flux_cmd("flux", no_errors=True),
            ),
            "local",
        ),
        "dry launch lsf": (
            f"{run} {lsf} --dry --local --no-errors --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR}",
            StepFileContainsCond(
                "runs", "*/runs.slurm.sh", "lsf_par", OUTPUT_DIR, "jsrun "
            ),
            "local",
        ),
        "dry launch slurm restart": (
            f"{run} {slurm_restart} --dry --local --no-errors --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR}",
            StepFileContainsCond(
                "runs",
                "*/runs.restart.slurm.sh",
                "slurm_par_restart",
                OUTPUT_DIR,
                "srun ",
            ),
            "local",
        ),
        "dry launch flux restart": (
            f"{run} {flux_restart} --dry --local --no-errors --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR}",
            StepFileContainsCond(
                "runs_rs",
                "*/runs_rs.restart.slurm.sh",
                "flux_par_restart",
                OUTPUT_DIR,
                get_flux_cmd("flux", no_errors=True),
            ),
            "local",
        ),
        "local override feature_demo": (
            f"{run} {demo} --vars N_SAMPLES=2 OUTPUT_PATH=./{OUTPUT_DIR} --local",
            [
                ReturnCodeCond(),
                ProvenanceCond(
                    regex=" -n 2 -outfile", name="feature_demo", output_path=OUTPUT_DIR
                ),
                StepFileExistsCond(
                    "verify", "MERLIN_FINISHED", "feature_demo", OUTPUT_DIR
                ),
            ],
            # "local",
        ),
        "local csv feature_demo": (
            f"echo 42.0,47.0 > foo_testing_temp.csv; {run} {demo} --samples foo_testing_temp.csv --vars OUTPUT_PATH=./{OUTPUT_DIR} --local; rm -f foo_testing_temp.csv",
            [RegexCond("1 sample loaded."), ReturnCodeCond()],
            # "local",
        ),
        "local tab feature_demo": (
            f"echo '42.0\t47.0\n7.0 5.3' > foo_testing_temp.tab; {run} {demo} --samples foo_testing_temp.tab --vars OUTPUT_PATH=./{OUTPUT_DIR} --local; rm -f foo_testing_temp.tab",
            [RegexCond("2 samples loaded."), ReturnCodeCond()],
            # "local",
        ),
        "distributed feature_demo": (
            f"{run} {demo} --vars OUTPUT_PATH=./{OUTPUT_DIR} WORKER_NAME=cli_test_demo_workers ; {workers} {demo} --vars OUTPUT_PATH=./{OUTPUT_DIR} WORKER_NAME=cli_test_demo_workers",
            [
                ReturnCodeCond(),
                ProvenanceCond(
                    regex="cli_test_demo_workers:",
                    name="feature_demo",
                    output_path=OUTPUT_DIR,
                ),
                StepFileExistsCond(
                    "verify", "MERLIN_FINISHED", "feature_demo", OUTPUT_DIR
                ),
            ],
        ),
        # "black check merlin": (f"{black} merlin/", ReturnCodeCond(), "local"),
        # "black check tests": (f"{black} tests/", ReturnCodeCond(), "local"),
        "deplic no GNU": (
            f"deplic ./",
            [RegexCond("GNU", negate=True), RegexCond("GPL", negate=True)],
            "local",
        ),
    }


def setup_argparse():
    parser = argparse.ArgumentParser(description="run_tests cli parser")
    parser.add_argument(
        "--exit",
        action="store_true",
        help="Flag for stopping all testing upon first failure",
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Flag for more detailed output messages"
    )
    parser.add_argument("--local", action="store_true", help="Run only local tests")
    parser.add_argument(
        "--ids",
        action="store",
        dest="ids",
        type=int,
        nargs="+",
        default=None,
        help="Provide space-delimited ids of tests you want to run."
        "Example: '--ids 1 5 8 13'",
    )
    return parser


def main():
    """
    High-level CLI test operations.
    """
    parser = setup_argparse()
    args = parser.parse_args()

    tests = define_tests()

    clear_test_studies_dir()
    result = run_tests(args, tests)
    clear_test_studies_dir()
    return result


if __name__ == "__main__":
    sys.exit(main())



import argparse

import matplotlib.pyplot as plt
import numpy as np
from joblib import dump
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error


descript = """Using parameters to edit OpenFOAM parameters"""
parser = argparse.ArgumentParser(description=descript)

parser.add_argument("-workspace", help="The DAG spec root")

args = parser.parse_args()

training_percent = 0.6
timestep = -1
fontsize = 20

WORKSPACE = args.workspace
inputs_dir = WORKSPACE + "/merlin_info"
outputs_dir = WORKSPACE + "/combine_outputs"

outputs = np.load(outputs_dir + "/data.npz")
U = outputs["arr_0"]
enstrophy = outputs["arr_1"]

energy_byhand = np.sum(np.sum(U ** 2, axis=3), axis=2) / U.shape[2] / 2
enstrophy_all = np.sum(enstrophy, axis=2)

X = np.load(inputs_dir + "/samples.npy")
y = np.concatenate(
    (
        enstrophy_all[:, timestep].reshape(-1, 1),
        energy_byhand[:, timestep].reshape(-1, 1),
    ),
    axis=1,
)
X[:, 1] = np.log10(X[:, 0] / X[:, 1])  # np.log10(X)
y = np.log10(y)

training_size = int(training_percent * len(X))

X_train = X[:training_size]
y_train = y[:training_size]
X_test = X[training_size:]
y_test = y[training_size:]

regr = RandomForestRegressor(max_depth=10, random_state=0, n_estimators=7)

regr.fit(X_train, y_train)
print("training score:", regr.score(X_train, y_train))
print("testing score: ", regr.score(X_test, y_test))
print(mean_squared_error(y_test, regr.predict(X_test)))

dump(regr, "trained_model.joblib")


fig, ax = plt.subplots(3, 2, figsize=(25, 25), constrained_layout=True)
plt.rcParams.update({"font.size": 25})
plt.rcParams["lines.linewidth"] = 5

x = np.linspace(-5, 8, 100)
y1 = 1 * x
ax[0][0].plot(x, y1, "-r", label="y=x", linewidth=1)

y_pred = regr.predict(X_train)
ax[0][0].scatter(y_train[:, 0], y_pred[:, 0], label="Log10 Enstrophy")
ax[0][0].scatter(y_train[:, 1], y_pred[:, 1], label="Log10 Energy")
ax[0][0].set_title("Velocity Magnitude %s" % timestep)

ax[0][0].set_xlabel("Actual", fontsize=fontsize)
ax[0][0].set_ylabel("Predicted", fontsize=fontsize)
ax[0][0].set_title("Training Data, # Points: %s" % len(y_pred))
ax[0][0].legend()
ax[0][0].grid()

x_min = np.min([np.min(y_train[:, 0]), np.min(y_train[:, 1])])
y_min = np.min([np.min(y_pred[:, 0]), np.min(y_pred[:, 1])])
x_max = np.max([np.max(y_train[:, 0]), np.max(y_train[:, 1])])
y_max = np.max([np.max(y_pred[:, 0]), np.max(y_pred[:, 1])])

y_pred = regr.predict(X_test)
ax[0][1].plot(x, y1, "-r", label="y=x", linewidth=1)
ax[0][1].scatter(y_test[:, 0], y_pred[:, 0], label="Log10 Enstrophy")
ax[0][1].scatter(y_test[:, 1], y_pred[:, 1], label="Log10 Energy")
ax[0][1].set_xlabel("Actual", fontsize=fontsize)
ax[0][1].set_ylabel("Predicted", fontsize=fontsize)
ax[0][1].set_title("Testing Data, # Points: %s" % len(y_pred))
ax[0][1].legend()
ax[0][1].grid()

x_min = np.min([np.min(y_test[:, 0]), np.min(y_test[:, 1]), x_min]) - 0.1
y_min = np.min([np.min(y_pred[:, 0]), np.min(y_pred[:, 1]), y_min]) - 0.1
x_max = np.max([np.max(y_test[:, 0]), np.max(y_test[:, 1]), x_max]) + 0.1
y_max = np.max([np.max(y_pred[:, 0]), np.max(y_pred[:, 1]), y_max]) + 0.1


ax[0][0].set_xlim([x_min, x_max])
ax[0][0].set_ylim([y_min, y_max])
ax[0][1].set_xlim([x_min, x_max])
ax[0][1].set_ylim([y_min, y_max])


y_pred_all = regr.predict(X)
input_enstrophy = ax[1][1].scatter(X[:, 0], 10 ** y[:, 1], s=100, edgecolors="black")
ax[1][1].set_xlabel(r"Lidspeed ($\frac{m}{s}$)", fontsize=fontsize)
ax[1][1].set_ylabel(r"$Energy$", fontsize=fontsize)
ax[1][1].set_title("Average Energy Variation with Lidspeed")
ax[1][1].grid()


input_energy = ax[1][0].scatter(
    X[:, 0],
    X[:, 1],
    s=100,
    edgecolors="black",
    c=10 ** y[:, 1],
    cmap=plt.get_cmap("viridis"),
)
ax[1][0].set_xlabel(r"Lidspeed ($\frac{m}{s}$)", fontsize=fontsize)
ax[1][0].set_ylabel(r"$Log_{10}$(Reynolds Number)", fontsize=fontsize)
ax[1][0].set_title("Inputs vs Average Energy")
ax[1][0].grid()
cbar = plt.colorbar(input_energy, ax=ax[1][0])
cbar.ax.set_ylabel(r"$Energy$", rotation=270, labelpad=30)

ax[1][0].tick_params(axis="both", which="major", labelsize=fontsize)
ax[1][1].tick_params(axis="both", which="major", labelsize=fontsize)
ax[1][0].tick_params(axis="both", which="major", labelsize=fontsize)
ax[1][1].tick_params(axis="both", which="major", labelsize=fontsize)


y_pred_all = regr.predict(X)
input_enstrophy = ax[2][0].scatter(
    X[:, 0],
    X[:, 1],
    s=100,
    edgecolors="black",
    c=y[:, 0] - y_pred_all[:, 0],
    cmap=plt.get_cmap("Spectral"),
)
ax[2][0].set_xlabel(r"Lidspeed ($\frac{m}{s}$)", fontsize=fontsize)
ax[2][0].set_ylabel(r"$Log_{10}$(Reynolds Number)", fontsize=fontsize)
ax[2][0].set_title("Inputs vs Enstrophy error")
ax[2][0].grid()
cbar = plt.colorbar(input_enstrophy, ax=ax[2][0])
cbar.ax.set_ylabel(r"$y_{act} - y_{pred}$", rotation=270, labelpad=30)


input_energy = ax[2][1].scatter(
    X[:, 0],
    X[:, 1],
    s=100,
    edgecolors="black",
    c=y[:, 1] - y_pred_all[:, 1],
    cmap=plt.get_cmap("Spectral"),
)
ax[2][1].set_xlabel(r"Lidspeed ($\frac{m}{s}$)", fontsize=fontsize)
ax[2][1].set_ylabel(r"$Log_{10}$(Reynolds Number)", fontsize=fontsize)
ax[2][1].set_title("Inputs vs Energy error")
ax[2][1].grid()
cbar = plt.colorbar(input_energy, ax=ax[2][1])
cbar.ax.set_ylabel(r"$y_{act} - y_{pred}$", rotation=270, labelpad=30)

ax[0][0].tick_params(axis="both", which="major", labelsize=fontsize)
ax[0][1].tick_params(axis="both", which="major", labelsize=fontsize)
ax[2][0].tick_params(axis="both", which="major", labelsize=fontsize)
ax[2][1].tick_params(axis="both", which="major", labelsize=fontsize)

plt.savefig("prediction.png")



"""Tests for the configfile module."""
import os
import shutil
import tempfile
import unittest
from getpass import getuser

from merlin.config import configfile

from .utils import mkfile


CONFIG_FILE_CONTENTS = """
celery:
    certs: path/to/celery/config/files

broker:
    name: rabbitmq
    username: testuser
    password: rabbit.password  # The filename that contains the password.
    server: jackalope.llnl.gov

results_backend:
    name: mysql
    dbname: testuser
    username: mlsi
    password: mysql.password  # The filename that contains the password.
    server: rabbit.llnl.gov

"""


class TestFindConfigFile(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.appfile = mkfile(self.tmpdir, "app.yaml")

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_tempdir(self):
        self.assertTrue(os.path.isdir(self.tmpdir))

    def test_find_config_file(self):
        """
        Given the path to a vaild config file, find and return the full
        filepath.
        """
        path = configfile.find_config_file(path=self.tmpdir)
        expected = os.path.join(self.tmpdir, self.appfile)
        self.assertEqual(path, expected)

    def test_find_config_file_error(self):
        """Given an invalid path, return None."""
        invalid = "invalid/path"
        expected = None

        path = configfile.find_config_file(path=invalid)
        self.assertEqual(path, expected)


class TestConfigFile(unittest.TestCase):
    """Unit tests for loading the config file."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.configfile = mkfile(self.tmpdir, "app.yaml", content=CONFIG_FILE_CONTENTS)

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_get_config(self):
        """
        Given the directory path to a valid merlin config file, then
        `get_config` should find the merlin config file and load the YAML
        contents to a dictionary.
        """
        expected = {
            "broker": {
                "name": "rabbitmq",
                "password": "rabbit.password",
                "server": "jackalope.llnl.gov",
                "username": "testuser",
                "vhost": getuser(),
            },
            "celery": {"certs": "path/to/celery/config/files"},
            "results_backend": {
                "dbname": "testuser",
                "name": "mysql",
                "password": "mysql.password",
                "server": "rabbit.llnl.gov",
                "username": "mlsi",
            },
        }

        self.assertDictEqual(configfile.get_config(self.tmpdir), expected)



import os
import shutil
import unittest
from contextlib import suppress

from merlin.common.sample_index_factory import (
    create_hierarchy,
    read_hierarchy,
)


TEST_DIR = "UNIT_TEST_SPACE"


def clear_test_tree():
    with suppress(FileNotFoundError):
        shutil.rmtree(TEST_DIR)


def clear(func):
    def wrapper():
        clear_test_tree()
        func()
        clear_test_tree()

    return wrapper


@clear
def test_index_file_writing():
    indx = create_hierarchy(
        1000000000, 10000, [100000000, 10000000, 1000000], root=TEST_DIR
    )
    indx.write_directories()
    indx.write_multiple_sample_index_files()
    indx2 = read_hierarchy(TEST_DIR)
    assert indx2.get_path_to_sample(123000123) == indx.get_path_to_sample(123000123)


def test_bundle_retrieval():
    indx = create_hierarchy(
        1000000000, 10000, [100000000, 10000000, 1000000], root=TEST_DIR
    )
    expected = f"{TEST_DIR}/0/0/0/samples0-10000.ext"
    result = indx.get_path_to_sample(123)
    assert expected == result

    expected = f"{TEST_DIR}/0/0/0/samples10000-20000.ext"
    result = indx.get_path_to_sample(10000)
    assert expected == result

    expected = f"{TEST_DIR}/1/2/3/samples123000000-123010000.ext"
    result = indx.get_path_to_sample(123000123)
    assert expected == result


def test_start_sample_id():
    expected = """: DIRECTORY MIN 203 MAX 303 NUM_BUNDLES 10
   0: BUNDLE 0 MIN 203 MAX 213
   1: BUNDLE 1 MIN 213 MAX 223
   2: BUNDLE 2 MIN 223 MAX 233
   3: BUNDLE 3 MIN 233 MAX 243
   4: BUNDLE 4 MIN 243 MAX 253
   5: BUNDLE 5 MIN 253 MAX 263
   6: BUNDLE 6 MIN 263 MAX 273
   7: BUNDLE 7 MIN 273 MAX 283
   8: BUNDLE 8 MIN 283 MAX 293
   9: BUNDLE 9 MIN 293 MAX 303
"""
    idx203 = create_hierarchy(100, 10, start_sample_id=203)
    assert expected == str(idx203)


@clear
def test_directory_writing():
    path = os.path.join(TEST_DIR)
    indx = create_hierarchy(2, 1, [1], root=path)
    expected = """: DIRECTORY MIN 0 MAX 2 NUM_BUNDLES 2
   0: DIRECTORY MIN 0 MAX 1 NUM_BUNDLES 1
      0.0: BUNDLE 0 MIN 0 MAX 1
   1: DIRECTORY MIN 1 MAX 2 NUM_BUNDLES 1
      1.0: BUNDLE 1 MIN 1 MAX 2
"""
    assert expected == str(indx)
    indx.write_directories()
    assert os.path.isdir(f"{TEST_DIR}/0")
    assert os.path.isdir(f"{TEST_DIR}/1")
    indx.write_multiple_sample_index_files()

    clear_test_tree()

    path = os.path.join(TEST_DIR)
    indx = create_hierarchy(1000000000, 10000, [100000000, 10000000], root=path)
    indx.write_directories()
    path = indx.get_path_to_sample(123000123)
    assert os.path.exists(os.path.dirname(path))
    assert path != TEST_DIR
    path = indx.get_path_to_sample(10000000000)
    assert path == TEST_DIR

    clear_test_tree()

    path = os.path.join(TEST_DIR)
    indx = create_hierarchy(
        1000000000, 10000, [100000000, 10000000, 1000000], root=path
    )
    indx.write_directories()


def test_directory_path():
    indx = create_hierarchy(20, 1, [20, 5, 1], root="")
    leaves = indx.make_directory_string()
    expected_leaves = "0/0/0 0/0/1 0/0/2 0/0/3 0/0/4 0/1/0 0/1/1 0/1/2 0/1/3 0/1/4 0/2/0 0/2/1 0/2/2 0/2/3 0/2/4 0/3/0 0/3/1 0/3/2 0/3/3 0/3/4"
    assert leaves == expected_leaves
    all_dirs = indx.make_directory_string(just_leaf_directories=False)
    expected_all_dirs = " 0 0/0 0/0/0 0/0/1 0/0/2 0/0/3 0/0/4 0/1 0/1/0 0/1/1 0/1/2 0/1/3 0/1/4 0/2 0/2/0 0/2/1 0/2/2 0/2/3 0/2/4 0/3 0/3/0 0/3/1 0/3/2 0/3/3 0/3/4"
    assert all_dirs == expected_all_dirs


@clear
def test_subhierarchy_insertion():
    indx = create_hierarchy(2, 1, [1], root=TEST_DIR)
    print("Writing directories")
    indx.write_directories()
    indx.write_multiple_sample_index_files()
    print("reading heirarchy")
    top = read_hierarchy(os.path.abspath(TEST_DIR))
    expected = """: DIRECTORY MIN 0 MAX 2 NUM_BUNDLES 2
   0: DIRECTORY MIN 0 MAX 1 NUM_BUNDLES 1
      0.0: BUNDLE -1 MIN 0 MAX 1
   1: DIRECTORY MIN 1 MAX 2 NUM_BUNDLES 1
      1.0: BUNDLE -1 MIN 1 MAX 2
"""
    assert str(top) == expected
    print("creating sub_heirarchy")
    sub_h = create_hierarchy(100, 10, address="1.0")
    print("inserting sub_heirarchy")
    top["1.0"] = sub_h
    print(str(indx))
    print("after insertion")
    print(str(top))
    expected = """: DIRECTORY MIN 0 MAX 2 NUM_BUNDLES 2
   0: DIRECTORY MIN 0 MAX 1 NUM_BUNDLES 1
      0.0: BUNDLE -1 MIN 0 MAX 1
   1: DIRECTORY MIN 1 MAX 2 NUM_BUNDLES 1
      1.0: DIRECTORY MIN 0 MAX 100 NUM_BUNDLES 10
         1.0.0: BUNDLE 0 MIN 0 MAX 10
         1.0.1: BUNDLE 1 MIN 10 MAX 20
         1.0.2: BUNDLE 2 MIN 20 MAX 30
         1.0.3: BUNDLE 3 MIN 30 MAX 40
         1.0.4: BUNDLE 4 MIN 40 MAX 50
         1.0.5: BUNDLE 5 MIN 50 MAX 60
         1.0.6: BUNDLE 6 MIN 60 MAX 70
         1.0.7: BUNDLE 7 MIN 70 MAX 80
         1.0.8: BUNDLE 8 MIN 80 MAX 90
         1.0.9: BUNDLE 9 MIN 90 MAX 100
"""
    assert str(top) == expected


def test_sample_index():
    """Run through some basic testing of the SampleIndex class."""
    tests = [
        (10, 1, []),
        (10, 3, []),
        (11, 2, [5]),
        (10, 3, [3]),
        (10, 3, [1]),
        (10, 1, [3]),
        (10, 3, [1, 3]),
        (10, 1, [2]),
        (1000, 100, [500]),
        (1000, 50, [500, 100]),
        (1000000000, 100000132, []),
    ]

    for args in tests:
        print(f"############ TEST {args[0]} {args[1]} {args[2]} ###########")
        # put at root address of "0" to guarantee insertion at "0.1" later is valid
        idx = create_hierarchy(args[0], args[1], args[2], address="0")
        print(str(idx))
        try:
            idx["0.1"] = create_hierarchy(args[0], args[1], args[2], address="0.1")
            print("successful set")
            print(str(idx))
        except KeyError as error:
            print(error)
            assert False



import os
import shutil
import tempfile
import unittest

from merlin.spec.specification import MerlinSpec


MERLIN_SPEC = """
description:
    name: basic_ensemble
    description: Run 100 hello worlds.

batch:
    type: local

env:
    variables:
        OUTPUT_PATH: ./studies
        HELLO: $(SPECROOT)/hello_world.py

study:
    - name: hello
      description: |
         process a sample with hello world
      run:
        cmd: |
          python $(HELLO) -outfile hello_world_output_$(merlin_sample_id).json $(X0) $(X1) $(X2)
        procs: 1
        nodes: 1
        task_queue: hello_queue

global.parameters:
    X2:
        values : [0.5]
        label  : X2.%%
    N_NEW:
        values : [100]
        label  : N_NEW.%%

merlin:
    samples:
        generate:
            cmd: python $(SPECROOT)/make_samples.py -n 100 -outfile=$(OUTPUT_PATH)/samples.npy
        file: $(OUTPUT_PATH)/samples.npy
        column_labels: [X0, X1]
"""


NO_MERLIN = """
description:
    name: basic_ensemble_no_merlin
    description: Run 100 hello worlds.

batch:
    type: local

env:
    variables:
        OUTPUT_PATH: ./studies
        HELLO: $(SPECROOT)/hello_world.py

study:
    - name: hello
      description: |
         process a sample with hello world
      run:
        cmd: |
          python $(HELLO) -outfile hello_world_output_$(merlin_sample_id).json $(X0) $(X1) $(X2)
        procs: 1
        nodes: 1
        task_queue: hello_queue

global.parameters:
    X2:
        values : [0.5]
        label  : X2.%%
    N_NEW:
        values : [100]
        label  : N_NEW.%%
"""


class TestMerlinSpec(unittest.TestCase):
    """Test the logic for parsing the Merlin spec into a MerlinSpec."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.merlin_spec_filepath = os.path.join(self.tmpdir, "basic_ensemble.yaml")

        with open(self.merlin_spec_filepath, "w+") as _file:
            _file.write(MERLIN_SPEC)

        self.spec = MerlinSpec.load_specification(self.merlin_spec_filepath)

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_specroot(self):
        """
        Given a Merlin spec with no other configurations to the MerlinSpec
        object, then the default spec root path should be set to the base
        path of the file.
        """
        self.assertEqual(self.spec.specroot, self.tmpdir)

    def test_filepath(self):
        """
        Given a Merlin spec with no other configurations to the MerlinSpec
        object, then the filepath should be set to the path of the yaml file 
        itself.
        """
        self.assertEqual(self.spec.path, self.merlin_spec_filepath)

    # TODO consider moving default behavior like this to MerlinStudy
    #    def test_task_server(self):
    #        """
    #        Given a Merlin spec with no Celery configuration specified in the spec,
    #        then the MerlinSpec object should default the task server to `celery`.
    #        """
    #        self.assertEqual(self.spec.merlin["resources"]["task_server, 'celery')

    def test_sample_files(self):
        """
        Given a Merlin spec with a sample file defined with a `SPECROOT`
        variable, when `spec.sample_files` is called, then the MerlinSpec
        object should return the expanded path (replaced SPECROOT) path to the
        sample files.
        """
        expected = "$(OUTPUT_PATH)/samples.npy"
        self.assertEqual(self.spec.merlin["samples"]["file"], expected)

    def test_sample_labels(self):
        """
        Given a Merlin spec with colume labels defined, when
        'spec.sample_labels' is called then the MerlinSpec object should
        return the values set in the column_labels.
        """
        self.assertEqual(self.spec.merlin["samples"]["column_labels"], ["X0", "X1"])

    def test_spec_sample_generator(self):
        """
        Given a Merlin spec with the generate section defined when
        'spec.sample_generate` is called then the MerlinSpec object should
        return the values set in the generate cmd section.
        """
        expected = "python $(SPECROOT)/make_samples.py -n 100 -outfile=$(OUTPUT_PATH)/samples.npy"
        self.assertEqual(self.spec.merlin["samples"]["generate"]["cmd"], expected)


class TestSpecNoMerlin(unittest.TestCase):
    """Test the logic for parsing a spec with no Merlin block into a MerlinSpec."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.merlin_spec_filepath = os.path.join(self.tmpdir, "no_merlin.yaml")

        with open(self.merlin_spec_filepath, "w+") as _file:
            _file.write(NO_MERLIN)

        self.spec = MerlinSpec.load_specification(self.merlin_spec_filepath)

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_default_merlin_block(self):
        self.assertEqual(self.spec.merlin["resources"]["task_server"], "celery")
        self.assertEqual(self.spec.merlin["resources"]["overlap"], False)
        self.assertEqual(
            self.spec.merlin["resources"]["workers"]["default_worker"]["steps"], ["all"]
        )
        self.assertEqual(
            self.spec.merlin["resources"]["workers"]["default_worker"]["batch"], None
        )
        self.assertEqual(
            self.spec.merlin["resources"]["workers"]["default_worker"]["nodes"], None
        )
        self.assertEqual(self.spec.merlin["samples"], None)



import argparse
import os
import sys

from concurrent.futures import ProcessPoolExecutor


def load_samples(sample_file_path):
    """Read in sample file, returning a flat list of all samples inside"""
    samples = []
    print("Reading: {}".format(sample_file_path))
    with open(os.path.abspath(sample_file_path), "r") as sample_file:
        for sample in sample_file:
            samples.append(sample.strip())

    return samples


def serialize_samples(sample_file_paths, outfile, nproc):
    """Writes out collection of samples, one entry per line"""
    with ProcessPoolExecutor(max_workers=nproc) as executor:
        all_samples = [
            sample
            for sample_list in executor.map(load_samples, sample_file_paths)
            for sample in sample_list
        ]

    with open(outfile, "w") as outfile:
        for sample in all_samples:
            outfile.write("{}\n".format(sample))


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="Read in samples from list of output files"
    )

    parser.add_argument(
        "sample_file_paths", help="paths to sample output files", default="", nargs="+"
    )

    parser.add_argument("--np", help="number of processors to use", type=int, default=1)

    parser.add_argument(
        "-outfile", help="Collected sample outputs", default="all_names.txt"
    )
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()

    # Collect sample files into single file
    sample_paths = [sample_path for sample_path in args.sample_file_paths]
    serialize_samples(sample_paths, args.outfile, args.np)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import os
import shutil
import subprocess


parser = argparse.ArgumentParser(description="Launch 35 merlin workflow jobs")
parser.add_argument("run_id", type=int, help="The ID of this run")
parser.add_argument("output_path", type=str, help="the output path")
parser.add_argument("spec_path", type=str, help="path to the spec to run")
args = parser.parse_args()

# launch 35 merlin workflow jobs
script_path = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
concurrencies = [1, 2, 4, 8, 16, 32, 64]
nodes = [1, 1, 1, 1, 1, 1, 2]
samples = [1, 10, 100, 1000, 10000]
output_path = os.path.join(args.output_path, f"run_{args.run_id}")
os.makedirs(output_path, exist_ok=True)
for i, concurrency in enumerate(concurrencies):
    c_name = os.path.join(output_path, f"c_{concurrency}")
    if not os.path.isdir(c_name):
        os.mkdir(c_name)
    os.chdir(c_name)
    for sample in samples:
        s_name = os.path.join(c_name, f"s_{sample}")
        if not os.path.isdir(s_name):
            os.mkdir(s_name)
        os.chdir(s_name)
        submit = f"submit_queue_tasks.sbatch"
        command = f"sbatch {submit} {sample} {int(concurrency/nodes[i])}"
        shutil.copyfile(os.path.join(script_path, submit), submit)
        shutil.copyfile(args.spec_path, os.path.basename(args.spec_path))
        lines = subprocess.check_output(command, shell=True).decode("ascii")
        os.chdir(f"..")
    os.chdir(f"..")



import argparse
import datetime
import re
import subprocess


# argument parsing
parser = argparse.ArgumentParser(description="Make some samples (names of people).")
parser.add_argument("logfile", type=str, help="celery log file")
parser.add_argument("errfile", type=str, help="batch err file")
parser.add_argument("c", type=int, help="concurrency")
parser.add_argument("s", type=int, help="n of samples")
args = parser.parse_args()


def single_task_times():
    pre_lines = subprocess.check_output(
        f'grep " succeeded in " {args.logfile}', shell=True
    ).decode("ascii")

    pre_list = pre_lines.strip().split("\n")

    task_durations = []
    for line in pre_list:
        matches = re.search(r"\d+.\d+s:", line)
        if matches:
            match = matches.group(0)
            match = float(match.strip("s:"))
            task_durations.append(match)

    print(str(task_durations))


def merlin_run_time():
    pre_line = subprocess.check_output(
        f'grep "real" {args.errfile}', shell=True
    ).decode("ascii")
    pre_line = pre_line.strip()
    matches = re.search(r"\d\.\d\d\d", pre_line)
    match = matches[0]
    result = float(match)
    print(f"c{args.c}_s{args.s} merlin run : " + str(result))


def start_verify_time():
    try:
        pre_line = subprocess.check_output(
            f'grep -m1 "verify" {args.logfile}', shell=True
        ).decode("ascii")
        pre_line = pre_line.strip()
        matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
        match = matches[0]
        element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
        timestamp = datetime.datetime.timestamp(element)
        print(f"c{args.c}_s{args.s} start verify : " + str(timestamp))
    except BaseException:
        print(f"c{args.c}_s{args.s} start verify : ERROR")


def start_run_workers_time():
    pre_line = subprocess.check_output(
        f'grep -m1 "" {args.logfile}', shell=True
    ).decode("ascii")
    pre_line = pre_line.strip()
    matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
    match = matches[0]
    element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
    timestamp = datetime.datetime.timestamp(element)
    print(f"c{args.c}_s{args.s} start run-workers : " + str(timestamp))


def start_sample1_time():
    pre_line = subprocess.check_output(
        f"grep -m1 \"Executing step 'null_step'\" {args.logfile}", shell=True
    ).decode("ascii")
    pre_line = pre_line.strip()
    matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
    match = matches[0]
    element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
    timestamp = datetime.datetime.timestamp(element)
    print(f"c{args.c}_s{args.s} start samp1 : " + str(timestamp))


def main():
    single_task_times()
    merlin_run_time()
    start_verify_time()
    start_run_workers_time()
    start_sample1_time()


if __name__ == "__main__":
    main()



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

#!/usr/bin/env python

"""
smallnpylib

A simple library to read the .npy file header and return a dict
containing the header from the .npy file as well as a few other
keys. Also provides the OpenNPY class which is a way of seeking
into .npy files without using the memory mapping functionality
in the load function. Finally, there is the OpenNPYList class
which opens a list of OpenNPY files and allows for random
access among all of them.

  'shape'         shape for the array
  'fortran_order' is it in fortran order (normally false)
  'descr'         the dtype description of the array
  'size'          total size of the data
  'offset'        position in file of start of data
  'rowsize'       the total size in bytes of a single row in the file (e.g. sample)
  'itemsize'      the size of a single element in the array

  SYNOPSIS :

    d = get_npy_info("myfile.txt");
    print d['offset']     # offset to data in file
    print d['rowsize']    # number of bytes per row

    with open("myfile.npy") as f :
      f.seek(d['offset']+d['rowsize]*N);

   OpenNPY

   with OpenNPY("myfile.npy") as a :
     print a[5]         # print row number 5
     print a[1:4]       # print rows from 1,2,3
     my_array = a.to_array();
     print len(a)       # number of rows in file
     for i in a :
       print i          # print all the rows in a
     print a.shape      # shape of array
     print a.dtype      # dtype of array

   with OpenNPYList(["myfile1.npy","myfile2.npy",...]) as a :
     print a[5]         # print row number 5
     print a[1:4]       # print rows from 1,2,3
     my_array = a.to_array();
     print len(a)       # number of rows in file
     for i in a :
       print i          # print all the rows in a
     print a.shape      # shape of array
     print a.dtype      # dtype of array

"""

import numpy as np


try:
    unistr = (unicode, str)
    npy_magic = "\x93NUMPY"
except NameError:
    unistr = str
    npy_magic = b"\x93NUMPY"


def _get_npy_info2(f):
    if isinstance(f, unistr):
        f = open(f, "rb")
    magic = f.read(6)  # must be .npy file
    assert magic == npy_magic  # must be .npy file or ELSE
    major, _ = list(map(ord, f.read(2)))
    if major == 1:
        hlen_char = list(map(ord, f.read(2)))
        hlen = hlen_char[0] + 256 * hlen_char[1]
    elif major == 2:
        hlen_char = list(map(ord, f.read(4)))
        hlen = (
            hlen_char[0]
            + 256 * hlen_char[1]
            + 65536 * hlen_char[2]
            + (1 << 24) * hlen_char[3]
        )
    else:
        raise Exception("unknown .npy format, e.g. not 1 or 2")
    hdr = eval(f.read(hlen))  # TODO remove eval
    hdr["dtype"] = np.dtype(hdr["descr"])
    hdr["offset"] = f.tell()  # location of data start
    hdr["itemsize"] = np.dtype(hdr["descr"]).itemsize
    hdr["rowsize"] = hdr["itemsize"] * np.product(hdr["shape"][1:])
    hdr["items"] = np.product(hdr["shape"])
    hdr["size"] = hdr["itemsize"] * hdr["items"]
    return f, hdr


def _get_npy_info3(f):
    if isinstance(f, unistr):
        f = open(f, "rb")
    magic = f.read(6)  # must be .npy file
    assert magic == npy_magic  # must be .npy file or ELSE
    major, _ = list(f.read(2))
    if major == 1:
        hlen_char = list(f.read(2))
        hlen = hlen_char[0] + 256 * hlen_char[1]
    elif major == 2:
        hlen_char = list(f.read(4))
        hlen = (
            hlen_char[0]
            + 256 * hlen_char[1]
            + 65536 * hlen_char[2]
            + (1 << 24) * hlen_char[3]
        )
    else:
        raise Exception("unknown .npy format, e.g. not 1 or 2")
    hdr = eval(f.read(hlen))  # TODO remove eval
    hdr["dtype"] = np.dtype(hdr["descr"])
    hdr["offset"] = f.tell()  # location of data start
    hdr["itemsize"] = np.dtype(hdr["descr"]).itemsize
    hdr["rowsize"] = hdr["itemsize"] * np.product(hdr["shape"][1:])
    hdr["items"] = np.product(hdr["shape"])
    hdr["size"] = hdr["itemsize"] * hdr["items"]
    return f, hdr


def _get_npy_info(f):
    d = None
    try:
        d = _get_npy_info2(f)
    except TypeError:
        d = _get_npy_info3(f)

    return d


def get_npy_info(f):
    try:
        d = _get_npy_info2(f)
    except TypeError:
        d = _get_npy_info3(f)
    d[0].close()
    return d[1]


def read_items(f, hdr, idx, n=-1, sep=""):
    f.seek(hdr["offset"] + idx * hdr["itemsize"])
    if n < 0:
        n = hdr["items"] - idx
    n = min(hdr["items"] - idx, n)
    print(f"n is {n}")
    return np.fromfile(f, dtype=hdr["dtype"], count=n, sep=sep)


def read_rows(f, hdr, idx, n=-1, sep=""):
    f.seek(hdr["offset"] + idx * hdr["rowsize"])
    if n < 0:
        n = hdr["shape"][0] - idx
    n = min(hdr["shape"][0] - idx, n)
    a = np.fromfile(
        f, dtype=hdr["dtype"], count=n * hdr["rowsize"] // hdr["itemsize"], sep=sep
    )
    return np.reshape(a, (n,) + hdr["shape"][1:])


def verify_open(func):  # A wrapper function used by FileSamples.
    """
      :param func: (function) a class instance method that needs to call
            _verify_open before doing anything else.
    """

    def wrapper(self, *v, **kw):
        self._verify_open()
        return func(self, *v, **kw)

    return wrapper


class OpenNPY:
    def __init__(self, f):
        self.hdr = self.f = None
        if isinstance(f, unistr):
            self.fname = f
        else:
            self.f = f
            self._verify_open()

    def _verify_open(self):
        if self.f is None:
            self.f, self.hdr = _get_npy_info(
                self.f if self.f is not None else self.fname
            )

    @verify_open
    def load_header(self, close=True):
        if close:
            self.close()
        return self.hdr

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tback):
        self.close()

    @verify_open
    def _shape(self):
        return self.hdr["shape"]

    shape = property(fget=_shape)

    @verify_open
    def _dtype(self):
        return self.hdr["dtype"]

    dtype = property(fget=_dtype)

    @verify_open
    def __getitem__(self, k):
        if isinstance(k, int):
            return read_rows(self.f, self.hdr, k, 1)[0]
        elif isinstance(k, slice):
            if k.step == 1 or k.step is None:
                return read_rows(self.f, self.hdr, k.start, k.stop - k.start)
            else:
                return np.asarray(
                    [
                        read_rows(self.f, self.hdr, _, 1)[0]
                        for _ in range(k.start, k.stop, 1 if k.step is None else k.step)
                    ]
                )

    @verify_open
    def __len__(self):
        return self.hdr["items"]

    @verify_open
    def __iter__(self):
        for k in range(self.hdr["shape"][0]):
            yield self[k]

    @verify_open
    def to_array(self):
        return read_rows(self.f, self.hdr, 0)


class OpenNPYList:
    def __init__(self, l):
        self.filenames = l
        self.files = [OpenNPY(_) for _ in self.filenames]
        for i in self.files:
            i.load_header()
        self.shapes = [_.hdr["shape"] for _ in self.files]
        for i in self.shapes[1:]:
            # Match subsequent axes shapes.
            assert i[1:] == self.shapes[0][1:]
        self.tells = np.cumsum([_[0] for _ in self.shapes])  # Tell locations.
        self.tells = np.hstack(([0], self.tells))

    def close(self):
        for i in self.files:
            i.close()

    def __del__(self):
        self.close()

    def __iter__(self):
        for i in self.files:
            for j in i:
                yield j

    def __getitem__(self, k):
        if isinstance(k, int):
            if k < 0:
                k = self.tells[-1] + k  # Negative indexing.
            if k >= self.tells[-1]:
                raise IndexError("index %d is out of bounds" % k)
            fno = (self.tells > k).argmax()
            return self.files[fno - 1][k - self.tells[fno - 1]]
        else:  # Slice indexing.
            # TODO : Implement a faster version.
            return np.asarray(
                [
                    self[_]
                    for _ in np.arange(
                        k.start, k.stop, k.step if k.step is not None else 1
                    )
                ]
            )

    def to_array(self):
        return np.vstack([_.to_array() for _ in self.files])

    def __len__(self):
        return self.tells[-1]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tback):
        self.close()


__all__ = [
    "OpenNPYList",
    "OpenNPY",
    "unistr",
    "read_items",
    "read_rows",
    "get_npy_info",
]

if __name__ == "__main__":
    import unittest
    import os
    import sys
    import uuid

    if "-h" in sys.argv:
        print(__doc__)
        sys.exit(1)

    class TestOpenNPY(unittest.TestCase):
        def test_seeknpy(self):
            e = np.diag(np.arange(5.0))
            try:
                fn = unicode(str(uuid.uuid1()) + ".npy")
            except NameError:
                fn = str(str(uuid.uuid1()) + ".npy")
            np.save(fn, e)
            with OpenNPY(fn) as a:
                ep = np.asarray([_ for _ in a])
            os.unlink(fn)
            self.assertTrue((ep == e).all())

        def test_seeknpylist(self):
            e = np.diag(np.arange(5.0))
            fn = str(uuid.uuid1()) + ".npy"
            np.save(fn, e)
            with OpenNPYList([fn, fn, fn]) as a:
                ep = np.asarray([_ for _ in a])
                en = np.asarray(a[5:10])
                en2 = np.asarray(a[11:14])
                en3 = np.asarray(a[1:14])
                en4 = a.to_array()
                # test __len__ method
                self.assertTrue(len(a) == 3 * len(e))
            os.unlink(fn)
            # test read slice of whole file
            self.assertTrue((en == e).all())
            self.assertTrue((en2 == e[1:4]).all())  # test slice
            # test read all
            self.assertTrue((ep == np.vstack((e, e, e))).all())
            # test to_array method
            self.assertTrue((en4 == np.vstack((e, e, e))).all())
            # test slice read across files
            self.assertTrue((en3 == np.vstack((e, e, e))[1:14]).all())

    unittest.main()



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
Manages formatting for displaying information to the console.
"""
import pprint
import subprocess

from kombu import Connection
from tabulate import tabulate

from merlin.ascii_art import banner_small
from merlin.config import (
    broker,
    results_backend,
)
from merlin.config.configfile import default_config_info


def check_server_access(sconf):
    servers = ["broker server", "results server"]

    if sconf.keys():
        print("\nChecking server connections:")
        print("-" * 28)

    excpts = {}
    for s in servers:
        if s in sconf:
            try:
                conn = Connection(sconf[s])
                conn.connect()
                conn.release()
                print(f"{s} connection: OK")
            except Exception as e:
                print(f"{s} connection: Error")
                excpts[s] = e

    if excpts:
        print("\nExceptions:")
        for k, v in excpts.items():
            print(f"{k}: {v}")


def display_config_info():
    """
    Prints useful configuration information to the console.
    """
    print("Merlin Configuration")
    print("-" * 25)
    print("")

    conf = default_config_info()
    sconf = {}
    excpts = {}
    try:
        conf["broker server"] = broker.get_connection_string(include_password=False)
        sconf["broker server"] = broker.get_connection_string()
        conf["broker ssl"] = broker.get_ssl_config()
    except Exception as e:
        conf["broker server"] = "Broker server error."
        excpts["broker server"] = e

    try:
        conf["results server"] = results_backend.get_connection_string(
            include_password=False
        )
        sconf["results server"] = results_backend.get_connection_string()
        conf["results ssl"] = results_backend.get_ssl_config()
    except Exception as e:
        conf["results server"] = "No results server configured or error."
        excpts["results server"] = e

    print(tabulate(conf.items(), tablefmt="presto"))

    if excpts:
        print("\nExceptions:")
        for k, v in excpts.items():
            print(f"{k}: {v}")

    check_server_access(sconf)


def display_multiple_configs(files, configs):
    """
    Logic for displaying multiple Merlin config files.

    :param `files`: List of merlin config files
    :param `configs`: List of merlin configurations
    """
    print("=" * 50)
    print(" MERLIN CONFIG ")
    print("=" * 50)

    for _file, config in zip(files, configs):
        print("")
        print(f"Display config info for path: {_file}")
        print("-" * 25)
        print("")
        pprint.pprint(config)


def print_info(args):
    """
    Provide version and location information about python and pip to
    facilitate user troubleshooting. 'merlin info' is a CLI tool only for
    developer versions of Merlin.

    :param `args`: parsed CLI arguments
    """
    print(banner_small)
    display_config_info()

    print("")
    print("Python Configuration")
    print("-" * 25)
    print("")
    info_calls = ["which python3", "python3 --version", "which pip3", "pip3 --version"]
    info_str = ""
    for x in info_calls:
        info_str += 'echo " $ ' + x + '" && ' + x + "\n"
        info_str += "echo \n"
    info_str += r"echo \"echo \$PYTHONPATH\" && echo $PYTHONPATH"
    subprocess.call(info_str, shell=True)
    print("")



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
This module parses the batch section of the yaml specification.

Currently only the batch worker launch for slurm, lsf or flux
are implemented.

"""
import logging
import os

from merlin.utils import get_yaml_var


LOG = logging.getLogger(__name__)


def batch_check_parallel(spec):
    """
    Check for a parallel batch section in the yaml file.
    """
    parallel = False

    try:
        batch = spec.batch
    except AttributeError:
        LOG.error("The batch section is required in the specification file.")
        raise

    btype = get_yaml_var(batch, "type", "local")
    if btype != "local":
        parallel = True

    return parallel


def get_batch_type(default=None):
    """
    Determine which batch scheduler to use.

    :param default: (str) The default batch scheduler to use if a scheduler
        can't be determined. The default is slurm.
    :returns: (str) The batch name (available options: slurm, flux, lsf).
    """
    if default is None:
        default = "slurm"

    if "SYS_TYPE" not in os.environ:
        return default

    if "toss3" in os.environ["SYS_TYPE"]:
        return "slurm"

    if "blueos" in os.environ["SYS_TYPE"]:
        return "lsf"

    return default


def get_node_count(default=1):
    """
    Determine a default node count based on the environment.

    :param default: (int) The number of nodes to return if a node count from
        the environment cannot be determined.
    :param returns: (int) The number of nodes to use.
    """
    if "SLURM_JOB_NUM_NODES" in os.environ:
        return int(os.environ["SLURM_JOB_NUM_NODES"])

    # LSB systems reserve one node for launching
    if "LSB_HOSTS" in os.environ:
        nodes = set(os.environ["LSB_HOSTS"].split())
        n_batch_nodes = len(nodes) - 1
        return n_batch_nodes
    elif "LSB_MCPU_HOSTS" in os.environ:
        nodes = os.environ["LSB_MCPU_HOSTS"].split()
        n_batch_nodes = len(nodes) // 2 - 1
        return n_batch_nodes

    return default


def batch_worker_launch(spec, com, nodes=None, batch=None):
    """
      The configuration in the batch section of the merlin spec
      is used to create the worker launch line, which may be
      different from a simulation launch.

      com (str): The command to launch with batch configuration
      nodes (int): The number of nodes to use in the batch launch
      batch (dict): An optional batch override from the worker config

    """
    if batch is None:
        try:
            batch = spec.batch
        except AttributeError:
            LOG.error("The batch section is required in the specification file.")
            raise

    btype = get_yaml_var(batch, "type", "local")

    # A jsrun submission cannot be run under a parent jsrun so
    # all non flux lsf submissions need to be local.
    if btype == "local" or "lsf" in btype:
        return com

    if nodes is None:
        # Use the value in the batch section
        nodes = get_yaml_var(batch, "nodes", None)

    # Get the number of nodes from the environment if unset
    if nodes is None or nodes == "all":
        nodes = get_node_count(default=1)

    # bank = get_yaml_var(batch, 'bank', "") #TODO this variable is unused
    queue = get_yaml_var(batch, "queue", "")
    shell = get_yaml_var(batch, "shell", "bash")
    launch_pre = get_yaml_var(batch, "launch_pre", "")
    launch_args = get_yaml_var(batch, "launch_args", "")
    worker_launch = get_yaml_var(batch, "worker_launch", "")

    if btype == "flux":
        launcher = get_batch_type()
    else:
        launcher = get_batch_type()

    launchs = worker_launch
    if not launchs:
        if btype == "slurm" or launcher == "slurm":
            launchs = f"srun --mpi=none -N {nodes} -n {nodes}"
            if queue:
                launchs += f" -p {queue}"
        if launcher == "lsf":
            launchs = f"jsrun -a 1 -c ALL_CPUS -g ALL_GPUS --bind=none -n {nodes}"

    launchs += f" {launch_args}"

    # Allow for any pre launch manipulation, e.g. module load
    # hwloc/1.11.10-cuda
    if launch_pre:
        launchs = f"{launch_pre} {launchs}"

    worker_cmd = f"{launchs} {com}"

    if btype == "flux":
        flux_path = get_yaml_var(batch, "flux_path", "")
        flux_opts = get_yaml_var(batch, "flux_start_opts", "")
        flux_exec_workers = get_yaml_var(batch, "flux_exec_workers", True)

        flux_exec = ""
        if flux_exec_workers:
            flux_exec = "flux exec"

        if "/" in flux_path:
            flux_path += "/"

        flux_exe = os.path.join(flux_path, "flux")

        launch = (
            f"{launchs} {flux_exe} start {flux_opts} {flux_exec} `which {shell}` -c"
        )
        worker_cmd = f'{launch} "{com}"'

    return worker_cmd



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
This module contains a list of examples that can be used when learning to use
Merlin, or for setting up new workflows.

Examples are packaged in directories, with the directory name denoting
the example name.  This must match the name of the merlin specification inside.
"""
import glob
import logging
import os
import shutil

import tabulate
import yaml


LOG = logging.getLogger(__name__)

EXAMPLES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "workflows")


def gather_example_dirs():
    result = {}
    for d in os.listdir(EXAMPLES_DIR):
        result[d] = d
    return result


def gather_all_examples():
    path = os.path.join(os.path.join(EXAMPLES_DIR, ""), os.path.join("*", "*.yaml"))
    return glob.glob(path)


def write_example(src_path, dst_path):
    """
    Write out the example workflow to a file.
    :param src_path: The path to copy from.
    :param content: The formatted content to write the file to.
    """
    if os.path.isdir(src_path):
        shutil.copytree(src_path, dst_path)
    else:
        shutil.copy(src_path, dst_path)


def list_examples():
    """List all available examples."""
    headers = ["name", "description"]
    rows = []
    for example_dir in gather_example_dirs():
        directory = os.path.join(os.path.join(EXAMPLES_DIR, example_dir), "")
        specs = glob.glob(directory + "*.yaml")
        for spec in specs:
            with open(spec) as f:
                try:
                    spec_metadata = yaml.safe_load(f)["description"]
                except KeyError:
                    LOG.warn(f"{spec} lacks required section 'description'")
                    continue
                except TypeError:
                    continue
            rows.append([spec_metadata["name"], spec_metadata["description"]])
    return "\n" + tabulate.tabulate(rows, headers) + "\n"


def setup_example(name, outdir):
    """Setup the given example."""
    example = None
    spec_paths = gather_all_examples()
    for spec_path in spec_paths:
        spec = os.path.basename(os.path.normpath(spec_path)).replace(".yaml", "")
        if name == spec:
            example = os.path.basename(os.path.dirname(spec_path))
            break
    if example is None:
        LOG.error(f"Example '{name}' not found.")
        return None

    # if there is only 1 file in the example, don't bother making a directory for it
    if len(os.listdir(os.path.dirname(spec_path))) == 1:
        src_path = os.path.join(EXAMPLES_DIR, os.path.join(example, example + ".yaml"))

    else:
        src_path = os.path.join(EXAMPLES_DIR, example)

        if outdir:
            outdir = os.path.join(os.getcwd(), outdir)
        else:
            outdir = os.path.join(os.getcwd(), example)

        if os.path.exists(outdir):
            LOG.error(f"File '{outdir}' already exists!")
            return None

    if outdir is None:
        outdir = os.getcwd()

    LOG.info(f"Copying example '{name}' to {outdir}")
    write_example(src_path, outdir)
    return example



"""
Tests for the maestroadapter.py module.
"""
import os
import shutil
import tempfile
import unittest

from merlin.study.step import Step
from merlin.study.study import MerlinStudy


MERLIN_SPEC = """
description:
    name: unit_test1
    description: Run 100 hello worlds.

batch:
    type: local

env:
    variables:
        OUTPUT_PATH: ./studies

    labels:
        SHARED: $(SPECROOT)/../shared
        HELLO: $(SHARED)/hello_world.py

study:
    - name: hello
      description: |
         process a sample with hello world
      run:
        cmd: |
          echo $PATH
          python $(HELLO) -outfile hello_world_output_$(merlin_sample_id).json $(X0) $(X1) $(X2)
        procs: 1
        nodes: 1
        task_queue: hello_queue

global.parameters:
    X2:
        values : [0.5]
        label  : X2.%%
    N_NEW:
        values : [100]
        label  : N_NEW.%%

merlin:
    samples:
        generate:
            cmd: python $(SPECROOT)/make_samples.py -n 100 -outfile=$(OUTPUT_PATH)/samples.npy
        file: $(OUTPUT_PATH)/samples.npy
        column_labels: [X0, X1]
"""


MERLIN_SPEC_NO_ENV = """
description:
    name: unit_test2
    description: Run 100 hello worlds.

batch:
    type: local

env:
    variables:
        OUTPUT_PATH: ./studies

    labels:
        SHARED: $(SPECROOT)/../shared
        HELLO: $(SHARED)/hello_world.py

study:
    - name: hello
      description: |
         process a sample with hello world
      run:
        cmd: |
          echo $PATH
          python $(HELLO) -outfile hello_world_output_$(merlin_sample_id).json $(X0) $(X1) $(X2)
        procs: 1
        nodes: 1
        task_queue: hello_queue

global.parameters:
    X2:
        values : [0.5]
        label  : X2.%%
    N_NEW:
        values : [100]
        label  : N_NEW.%%

merlin:
    samples:
        generate:
            cmd: python $(SPECROOT)/make_samples.py -n 100 -outfile=$(OUTPUT_PATH)/samples.npy
        file: $(OUTPUT_PATH)/samples.npy
        column_labels: [X0, X1]
"""

MERLIN_SPEC_CONFLICT = """
description:
    name: unit_test3
    description: Run 100 hello worlds.

batch:
    type: local

env:
    variables:
        OUTPUT_PATH: ./studies

    labels:
        SHARED: $(SPECROOT)/../shared
        HELLO: $(SHARED)/hello_world.py

study:
    - name: hello
      description: |
         process a sample with hello world
      run:
        cmd: |
          echo $PATH
          python $(HELLO) -outfile hello_world_output_$(merlin_sample_id).json $(X0) $(X1) $(X2)
        procs: 1
        nodes: 1
        task_queue: hello_queue

global.parameters:
    X2:
        values : [0.5]
        label  : X2.%%
    N_NEW:
        values : [100]
        label  : N_NEW.%%
    X1:
        values : [0.5]
        label  : X1.%%

merlin:
    samples:
        generate:
            cmd: python $(SPECROOT)/make_samples.py -n 100 -outfile=$(OUTPUT_PATH)/samples.npy
        file: $(OUTPUT_PATH)/samples.npy
        column_labels: [X0, X1]
"""


def test_get_task_queue_default():
    """
    Given a steps dictionary that sets the task queue to `test_queue` return
    `test_queue` as the queue name.
    """
    steps = {"run": {"task_queue": "test_queue"}}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "test_queue"


def test_get_task_queue_task_queue_missing():
    """
    Given a steps dictionary  where the run is set to an empty dictionary
    return `merlin` as the queue name.
    """
    steps = {"run": {}}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_get_task_queue_run_missing():
    """
    Given an empty steps dictionary return `merlin` as the queue name.
    """
    steps = {}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_get_task_queue_steps_None():
    """
    Given the value of None return `merlin` as the queue name.
    """
    steps = None
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_get_task_queue_run_None():
    """
    Given a steps dictionary where the run value is set to None, return
    `merlin` as the queue name.
    """
    steps = {"run": None}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_get_task_queue_None():
    """
    Given a steps dictionary where the task_queue is set to None, return
    `merlin` as the queue name.
    """
    steps = {"run": {"task_queue": None}}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_mastro_task_queue_None_str():
    """
    Given a steps dictionary where the task_queue is set to the string value
    'None`, return `merlin` as the queue name.
    """
    steps = {"run": {"task_queue": "None"}}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


def test_get_task_queue_none_str():
    """
    Given a steps dictionary where the task_queue is set to the string value
    'none', return `merlin` as the queue name.
    """
    steps = {"run": {"task_queue": "none"}}
    queue = Step.get_task_queue_from_dict(steps)
    assert queue == "merlin"


class TestMerlinStudy(unittest.TestCase):
    """Test the logic for parsing the MerlinStudy."""

    @staticmethod
    def file_contains_string(f, string):
        return string in open(f, "r").read()

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.merlin_spec_filepath = os.path.join(self.tmpdir, "basic_ensemble.yaml")

        with open(self.merlin_spec_filepath, "w+") as _file:
            _file.write(MERLIN_SPEC)

        self.study = MerlinStudy(self.merlin_spec_filepath)

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_expanded_spec(self):
        """
        Given a Merlin spec with no other configurations to the MerlinSpec
        object, the MerlinStudy should produce a new spec with all instances
        of $(OUTPUT_PATH), $(SPECROOT), and env labels and variables expanded. 
        """
        assert TestMerlinStudy.file_contains_string(
            self.merlin_spec_filepath, "$(SPECROOT)"
        )
        assert TestMerlinStudy.file_contains_string(
            self.merlin_spec_filepath, "$(OUTPUT_PATH)"
        )
        assert TestMerlinStudy.file_contains_string(self.merlin_spec_filepath, "$PATH")

        assert not TestMerlinStudy.file_contains_string(
            self.study.expanded_spec.path, "$(SPECROOT)"
        )
        assert not TestMerlinStudy.file_contains_string(
            self.study.expanded_spec.path, "$(OUTPUT_PATH)"
        )
        assert not TestMerlinStudy.file_contains_string(
            self.study.expanded_spec.path, "$PATH"
        )

    def test_column_label_conflict(self):
        """
        If there is a common key between Maestro's global.parameters and
        Merlin's sample/column_labels, an error should be raised.
        """
        merlin_spec_conflict = os.path.join(self.tmpdir, "basic_ensemble_conflict.yaml")
        with open(merlin_spec_conflict, "w+") as _file:
            _file.write(MERLIN_SPEC_CONFLICT)
        try:
            study_conflict = MerlinStudy(merlin_spec_conflict)
        except ValueError:
            pass
        else:
            assert False

    def test_no_env(self):
        """
        A MerlinStudy should be able to support a MerlinSpec that does not contain
        the optional `env` section.
        """
        merlin_spec_no_env_filepath = os.path.join(
            self.tmpdir, "basic_ensemble_no_env.yaml"
        )
        with open(merlin_spec_no_env_filepath, "w+") as _file:
            _file.write(MERLIN_SPEC_NO_ENV)
        try:
            study_no_env = MerlinStudy(merlin_spec_no_env_filepath)
        except Exception as e:
            assert False



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.5.2.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
Holds DAG class. TODO make this an interface, separate from Maestro.
"""
from collections import OrderedDict

from merlin.study.step import Step


class DAG:
    """
    This class provides methods on a task graph that Merlin needs for staging
    tasks in celery. It is initialized from am maestro ExecutionGraph, and the
    major entry point is the group_tasks method, which provides groups of
    independent chains of tasks.
    """

    def __init__(self, maestro_dag, labels):
        """
        :param `maestro_dag`: A maestrowf ExecutionGraph.
        """
        self.dag = maestro_dag
        self.backwards_adjacency = {}
        self.calc_backwards_adjacency()
        self.labels = labels

    def step(self, task_name):
        """Return a Step object for the given task name

        :param `task_name`: The task name.
        :return: A Merlin Step object.
        """
        return Step(self.dag.values[task_name])

    def calc_depth(self, node, depths, current_depth=0):
        """Calculate the depth of the given node and its children.

        :param `node`: The node (str) to start at.
        :param `depths`: the dictionary of depths to update.
        :param `current_depth`: the current depth in the graph traversal.
        """
        if node not in depths:
            depths[node] = current_depth
        else:
            depths[node] = max(depths[node], current_depth)

        for child in self.children(node):
            self.calc_depth(child, depths, current_depth=depths[node] + 1)

    @staticmethod
    def group_by_depth(depths):
        """Group DAG tasks by depth.

        :param `depths`: the dictionary of depths to group by

        :return: a list of lists of lists ordered by depth

        ([[["tasks"],["with"],["Depth 0"]],[["tasks"],["with"],["Depth 1"]]])

        The outer index of this list is the depth, the middle index is which
        chain of tasks in that depth, and the inner index is the task id in
        that chain.

        This method only groups by depth, and has one task in every chain.
        find_independent_chains is used to figure out how to coalesce chains
        across depths.
        """
        groups = {}
        for node in depths:
            depth = depths[node]

            if depth not in groups:
                groups[depth] = [node]
            else:
                groups[depth].append(node)

        # return an dict ordered by depth
        ordered_groups = OrderedDict(sorted(groups.items(), key=lambda t: t[0]))

        list_of_groups_of_chains = [[[g] for g in x] for x in ordered_groups.values()]

        return list_of_groups_of_chains

    def children(self, task_name):
        """ Return the children of the task.
        :param `task_name`: The name of the task to get the children of.

        :return: list of children of this task.
        """
        return self.dag.adjacency_table[task_name]

    def num_children(self, task_name):
        """ Find the number of children for the given task in the dag.
        :param `task_name`: The name of the task to count the children of.

        :return : number of children this task has
        """
        return len(self.children(task_name))

    def parents(self, task_name):
        """ Return the parents of the task.
        :param `task_name` : The name of the task to get the parents of.

        :return : list of parents of this task"""
        return self.backwards_adjacency[task_name]

    def num_parents(self, task_name):
        """find the number of parents for the given task in the dag
        :param `task_name` : The name of the task to count the parents of

        :return : number of parents this task has"""
        return len(self.parents(task_name))

    @staticmethod
    def find_chain(task_name, list_of_groups_of_chains):
        """find the chain containing the task
        :param `task_name` : The task to search for.
        :param `list_of_groups_of_chains` : list of groups of chains to search
            for the task

        :return : the list representing the chain containing task_name"""
        for group in list_of_groups_of_chains:
            for chain in group:
                if task_name in chain:
                    return chain
        return None

    def calc_backwards_adjacency(self):
        """ initializes our backwards adjacency table """
        for parent in self.dag.adjacency_table:
            for task_name in self.dag.adjacency_table[parent]:
                if task_name in self.backwards_adjacency:
                    self.backwards_adjacency[task_name].append(parent)
                else:
                    self.backwards_adjacency[task_name] = [parent]

    def compatible_merlin_expansion(self, task1, task2):
        """
        TODO
        """
        step1 = self.step(task1)
        step2 = self.step(task2)
        return step1.needs_merlin_expansion(
            self.labels
        ) == step2.needs_merlin_expansion(self.labels)

    def find_independent_chains(self, list_of_groups_of_chains):
        """
        Finds independent chains and adjusts with the groups of chains to
        maximalize parallelism

        :param list_of_groups_of_chains: List of list of lists, as returned by
            self.group_by_depth

        e.g.,

        ([[["task1"],["with"],["Depth 0"]],[["task2"],["has"],["Depth 1"]]])

        :return : This takes the groups of chains and looks for opportunities
            to move tasks in deeper groups of chains into chains in shallower
            groups, thus increasing available parallelism in the execution.

            Depending on the precise parental relationships between the tasks
            in the graph the output may be something like:

            ([[["task1", "has"],["with","task2"],["Depth 0"]],["Depth 1"]]])
        """
        for group in list_of_groups_of_chains:
            for chain in group:
                for task_name in chain:

                    if self.num_children(task_name) == 1 and task_name != "_source":

                        child = self.children(task_name)[0]

                        if self.num_parents(child) == 1:

                            if self.compatible_merlin_expansion(child, task_name):

                                self.find_chain(child, list_of_groups_of_chains).remove(
                                    child
                                )

                                chain.append(child)

        new_list = [
            [chain for chain in group if len(chain) > 0]
            for group in list_of_groups_of_chains
        ]
        new_list_2 = [group for group in new_list if len(group) > 0]

        return new_list_2

    def group_tasks(self, source_node):
        """Group independent tasks in a directed acyclic graph (DAG).

        Starts from a source node and works down, grouping tasks by
        depth, then identify independent parallel chains in those groups.

        :param dag : The DAG
        :param source_node: The source node.
        """
        depths = {}
        self.calc_depth(source_node, depths)
        groups_of_chains = self.group_by_depth(depths)

        return self.find_independent_chains(groups_of_chains)



"""Tests for the results_backend module."""
import os
import shutil
import tempfile
import unittest

from merlin.config import results_backend

from .utils import mkfile


class TestResultsBackend(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

        # Create test files.
        self.tmpfile1 = mkfile(self.tmpdir, "mysql_test1.txt")
        self.tmpfile2 = mkfile(self.tmpdir, "mysql_test2.txt")

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_mysql_config(self):
        """
        Given the path to a directory containing the MySQL cert files and a
        dictionary of files to look for, then find and return the full path to
        all the certs.
        """
        certs = {"test1": "mysql_test1.txt", "test2": "mysql_test2.txt"}

        # This will just be the above dictionary with the full file paths.
        expected = {
            "test1": os.path.join(self.tmpdir, certs["test1"]),
            "test2": os.path.join(self.tmpdir, certs["test2"]),
        }
        results = results_backend.get_mysql_config(self.tmpdir, certs)
        self.assertDictEqual(results, expected)

    def test_mysql_config_no_files(self):
        """
        Given the path to a directory containing the MySQL cert files and
        an empty dictionary, then `get_mysql_config` should return an empty
        dictionary.
        """
        files = {}
        result = results_backend.get_mysql_config(self.tmpdir, files)
        self.assertEqual(result, {})


class TestConfingMysqlErrorPath(unittest.TestCase):
    """
    Test `get_mysql_config` against cases were the given path does not exist.
    """

    def test_mysql_config_false(self):
        """
        Given a path that does not exist, then `get_mysql_config` should return
        False.
        """
        path = "invalid/path"

        # We don't need the dictionary populated for this test. The function
        # should return False before trying to process the dictionary.
        certs = {}
        result = results_backend.get_mysql_config(path, certs)
        self.assertFalse(result)
